/////////////////////////////////////////////////////////////////////////////
// GliD3D_Buffer.cpp : No comment
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Directive

#include "StdHdr.h"
#pragma hdrstop
#include "GliD3D.h"

/////////////////////////////////////////////////////////////////////////////
// Defines / Variables

/////////////////////////////////////////////////////////////////////////////
// Prototypes

/////////////////////////////////////////////////////////////////////////////
// Implements

FX_ENTRY void FX_CALL
grBufferClear( GrColor_t color, GrAlpha_t alpha, FxU16 depth )
{
	TRACEFUNCTION("grBufferClear");

	DWORD cr = grsubConvColor(color);
	switch (g_gr_state.depth_buffer_mode) {
	case GR_DEPTHBUFFER_DISABLE:
		// Clear back buffer.
		GB3DClear(cr, 1.0f, GB3DCLEAR_TARGET);
		break;
	case GR_DEPTHBUFFER_WBUFFER:
		// Clear back buffer and W-buffer.
		GB3DClear(cr, (FLOAT)(depth) / (FLOAT)GR_WDEPTHVALUE_FARTHEST, GB3DCLEAR_TARGET | GB3DCLEAR_ZBUFFER);
		break;
	case GR_DEPTHBUFFER_ZBUFFER:
		// Clear back buffer and Z-buffer.
		GB3DClear(cr, (FLOAT)(depth) / (FLOAT)GR_ZDEPTHVALUE_NEAREST, GB3DCLEAR_TARGET | GB3DCLEAR_ZBUFFER);
		break;
	}
}

FX_ENTRY int FX_CALL
grBufferNumPending( void )
{
	TRACEFUNCTION("grBufferNumPending");

	// First of all return 0.

	return 0;
}

FX_ENTRY void FX_CALL
grBufferSwap( int swap_interval )
{
	TRACEFUNCTION("grBufferSwap");
	GB3DEndRender();

#ifdef GLID3D_SHOWDEBUGINFO

	// Show Triangles Per Frame and Frames Per Second.
	static FLOAT fFPS = 0;
	static INT nNowTick = 0, nOldTick = 0, nNextTick = 0;

	nNowTick = GetTickCount();
	if (nNowTick > nNextTick) {
		// In 200 millisecond interval.
		nNextTick = nNowTick + 200;
		if (nOldTick != 0) {
			fFPS = 1000.0f / (nNowTick - nOldTick);
		}
	}
	nOldTick = nNowTick;

	// If debug build then show TPF and FPS else FPS only.
	#ifdef _DEBUG

		CHAR pszText[128];
		wsprintf(pszText, "Triangles per frame = %d", g_gr_nTrianglePerFrame);
		GB2DTextOut(-1.0f, -1.0f, pszText);
		sprintf(pszText, "Frames per second = %6.2f", fFPS);
		GB2DTextOut(-1.0f, -0.9f, pszText);

		g_gr_nTrianglePerFrame = 0;

	#else

		CHAR pszText[128];
		sprintf(pszText, "Frames per second = %6.2f", fFPS);
		GB2DTextOut(-1.0f, -1.0f, pszText);

	#endif

#endif

	// Below code, I try RED & BLUE glass mode. But not yet work.
/*
	if (!g_gr_bNoGammaControl && g_gr_bStereo) {

		DDGAMMARAMP ddgammaramp;
		INT i, nValue;
		for (i = 0; i < 256; i++) {
			if (g_gr_bLeft) {
				nValue = (DWORD)((FLOAT)i * (FLOAT)g_gr_fGamma * (FLOAT)0xffff / 255.0f);
				if (nValue < 0) nValue =0; else if (nValue > 0xffff) nValue = 0xffff;
				ddgammaramp.red[i] = (WORD)nValue;
				nValue = (DWORD)((FLOAT)i * (FLOAT)g_gr_fGamma * (FLOAT)0x0000 / 255.0f);
				if (nValue < 0) nValue =0; else if (nValue > 0xffff) nValue = 0xffff;
				ddgammaramp.green[i] = (WORD)nValue;
				nValue = (DWORD)((FLOAT)i * (FLOAT)g_gr_fGamma * (FLOAT)0x8000 / 255.0f);
				if (nValue < 0) nValue =0; else if (nValue > 0xffff) nValue = 0xffff;
				ddgammaramp.blue[i] = (WORD)nValue;
			} else {
				nValue = (DWORD)((FLOAT)i * (FLOAT)g_gr_fGamma * (FLOAT)0x0000 / 255.0f);
				if (nValue < 0) nValue =0; else if (nValue > 0xffff) nValue = 0xffff;
				ddgammaramp.red[i] = (WORD)nValue;
				nValue = (DWORD)((FLOAT)i * (FLOAT)g_gr_fGamma * (FLOAT)0xffff / 255.0f);
				if (nValue < 0) nValue =0; else if (nValue > 0xffff) nValue = 0xffff;
				ddgammaramp.green[i] = (WORD)nValue;
				nValue = (DWORD)((FLOAT)i * (FLOAT)g_gr_fGamma * (FLOAT)0x8000 / 255.0f);
				if (nValue < 0) nValue =0; else if (nValue > 0xffff) nValue = 0xffff;
				ddgammaramp.blue[i] = (WORD)nValue;
			}
		}

		HRESULT hr;
		LPDIRECTDRAWGAMMACONTROL pddgamma = NULL;
		hr = g_gr_pddsPrim->QueryInterface(IID_IDirectDrawGammaControl, (VOID**)&pddgamma);
		if (SUCCEEDED(hr)) {
			pddgamma->SetGammaRamp(0, &ddgammaramp);
			RELEASE(pddgamma);
		}

//		static nWait = 0;
//		if (nWait++ > 10) {
			g_gr_bLeft = !g_gr_bLeft;
//			nWait = 0;
//		}
	}
*/
	GBVideoShow();
	GB3DBeginRender();
}

FX_ENTRY void FX_CALL
grRenderBuffer( GrBuffer_t buffer )
{
	TRACEFUNCTION("grRenderBuffer");

	g_gr_state.render_buffer = buffer;
	switch (buffer) {
	case GR_BUFFER_FRONTBUFFER: GB3DSetRenderTarget(GBVIDEO_FRONTBUFFER); break;
	case GR_BUFFER_BACKBUFFER:  GB3DSetRenderTarget(GBVIDEO_BACKBUFFER); break;
	}
}

/////////////////////////////////////////////////////////////////////////////
