/////////////////////////////////////////////////////////////////////////////
// GliD3D_Render.cpp : No comment
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Directive

#include "StdHdr.h"
#pragma hdrstop
#include "GliD3D.h"

/////////////////////////////////////////////////////////////////////////////
// Defines / Variables

#define GLID3D_ANTIALIAS_ENABLE \
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ANTIALIAS, D3DANTIALIAS_SORTINDEPENDENT); \
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_EDGEANTIALIAS, TRUE);

#define GLID3D_ANTIALIAS_DISABLE \
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ANTIALIAS, D3DANTIALIAS_NONE); \
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_EDGEANTIALIAS, FALSE);

//#define GLID3D_D3DDP_DONOTCLIP	(D3DDP_DONOTCLIP)
#define GLID3D_D3DDP_DONOTCLIP	(0)

static D3DTLVERTEX tlv[4096];
static WORD idx[4096];
static FLOAT f256 = 256.0f;
static FLOAT f4096 = 4096.0f;
static FLOAT f65535 = 65535.0f;
static FLOAT f65536 = 65536.0f;

/////////////////////////////////////////////////////////////////////////////
// Prototypes

/////////////////////////////////////////////////////////////////////////////
// Implements

inline VOID
grsubSetTLVertexFromGrVertex(LPD3DTLVERTEX ptlv, const GrVertex* pv)
{

#ifndef GLID3D_NOSCALINGFIX

	if (g_gr_bScaleFix) {
		if (pv->x < f4096)
			ptlv->sx = pv->x * g_gr_fWidthScale;
		else
			ptlv->sx = (pv->x - (FLOAT)(((DWORD)pv->x) & 0xfffff000)) * g_gr_fWidthScale;

		if (g_gr_state.origin_location == GR_ORIGIN_UPPER_LEFT) {
			if (pv->y < f4096)
				ptlv->sy = pv->y * g_gr_fHeightScale;
			else
				ptlv->sy = (pv->y - (FLOAT)(((DWORD)pv->y) & 0xfffff000)) * g_gr_fHeightScale;

		} else {
			if (pv->y < f4096)
				ptlv->sy = ((FLOAT)g_gr_nHeight - pv->y) * g_gr_fHeightScale;
			else
				ptlv->sy = ((FLOAT)g_gr_nHeight - (FLOAT)(((DWORD)pv->y) & 0xfffff000)) * g_gr_fHeightScale;

		}
	} else {

#endif

		if (pv->x < f4096)
			ptlv->sx = pv->x;
		else
			ptlv->sx = pv->x - (FLOAT)(((DWORD)pv->x) & 0xfffff000);

		if (g_gr_state.origin_location == GR_ORIGIN_UPPER_LEFT) {
			if (pv->y < f4096)
				ptlv->sy = pv->y;
			else
				ptlv->sy = pv->y - (FLOAT)(((DWORD)pv->y) & 0xfffff000);

		} else {
			if (pv->y < f4096)
				ptlv->sy = (FLOAT)g_gr_nHeight - pv->y;
			else
				ptlv->sy = (FLOAT)g_gr_nHeight - (FLOAT)(((DWORD)pv->y) & 0xfffff000);

		}

#ifndef GLID3D_NOSCALINGFIX

	}

#endif

	switch (g_gr_state.depth_buffer_mode) {
	case GR_DEPTHBUFFER_ZBUFFER:
		ptlv->sz = f65535 / pv->ooz;
		ptlv->rhw = 1.0f / ptlv->sz;
		break;
	case GR_DEPTHBUFFER_WBUFFER:
		if (!g_gr_bUseZForW) {
			ptlv->sz = 1.0f - pv->oow;
			ptlv->rhw = pv->oow;
		} else {
			ptlv->sz = 1.0f / pv->oow / f65535;
			ptlv->rhw = 1.0f / ptlv->sz;
		}
		break;
	case GR_DEPTHBUFFER_DISABLE:
		ptlv->sz = f65535 / pv->ooz;
		ptlv->rhw = 1.0f / ptlv->sz;
		break;
	}

//	if (g_gr_bStereo) {
//
//		if (g_gr_state.depth_buffer_mode == GR_DEPTHBUFFER_ZBUFFER) {
//			ptlv->sx += (10.0f * ptlv->sz * ptlv->sz * ptlv->sz) * (g_gr_bLeft ? 1.0f : -1.0f);
//		} else if (g_gr_state.depth_buffer_mode == GR_DEPTHBUFFER_WBUFFER) {
//			ptlv->sx += (10.0f * (1.0f - ptlv->rhw * ptlv->rhw * ptlv->rhw)) * (g_gr_bLeft ? 1.0f : -1.0f);
//		}
//
//	}	
	
	if (g_gr_state.alpha_blend_function_rgb_sf != GR_BLEND_PREFOG_COLOR) {

		switch (g_gr_state.color_combine_local) {
		case GR_COMBINE_LOCAL_ITERATED:
			switch (g_gr_state.alpha_combine_local) {
			case GR_COMBINE_LOCAL_ITERATED:
//				ptlv->color = ((DWORD)pv->r << 16) | ((DWORD)pv->g <<  8) | ((DWORD)pv->b <<  0) | ((DWORD)pv->a << 24);
				ptlv->color = ((((DWORD)pv->a << 8) | ((DWORD)pv->r)) << 16) | ((DWORD)pv->g <<  8) | ((DWORD)pv->b);
//				ptlv->color = (DWORD)(pv->a * 255.0f * 65536.0f + pv->r * 65536.0f + pv->g * 255.0f + pv->b);
				break;
			case GR_COMBINE_LOCAL_CONSTANT:
				ptlv->color = ((DWORD)pv->r << 16) | ((DWORD)pv->g <<  8) | ((DWORD)pv->b <<  0) | (g_gr_state.constant_color_value_alpha);
//				ptlv->color = (DWORD)(pv->r * 65536.0f + pv->g * 255.0f + pv->b) + g_gr_state.constant_color_value_alpha;
				break;
			}
			break;
		case GR_COMBINE_LOCAL_CONSTANT:
			switch (g_gr_state.alpha_combine_local) {
			case GR_COMBINE_LOCAL_ITERATED:
				ptlv->color = g_gr_state.constant_color_value_color | ((DWORD)pv->a << 24);
				break;
			case GR_COMBINE_LOCAL_CONSTANT:
				ptlv->color = g_gr_state.constant_color_value;
				break;
			}
			break;
		}
	} else {
//		ptlv->color = g_gr_state.fog_d3dcolor;
	}
	switch (g_gr_state.color_combine_function) {
	case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA:
//		FLOAT fF;
//		fF = pv->a + pv->a * 255.0f;
//		fF = fF + fF * 65536.0f;
//		ptlv->specular = (DWORD)fF;
		ptlv->specular = ((DWORD)pv->a) | ((DWORD)pv->a << 8);
		ptlv->specular |= ptlv->specular << 16;
		break;
	default:
		ptlv->specular = 0;
	}
	ptlv->tu = ((pv->tmuvtx[0].sow / pv->oow) + 0.2f) / f256;
	ptlv->tv = ((pv->tmuvtx[0].tow / pv->oow) + 0.2f) / f256;
//	if (g_gr_tmu[0].s_clampmode == GR_TEXTURECLAMP_CLAMP) {
//		if (ptlv->tu < 0.0f) ptlv->tu = 0.0f;
//		if (ptlv->tu > 1.0f) ptlv->tu = 1.0f;
//	}
//	if (g_gr_tmu[0].t_clampmode == GR_TEXTURECLAMP_CLAMP) {
//		if (ptlv->tv < 0.0f) ptlv->tv = 0.0f;
//		if (ptlv->tv > 1.0f) ptlv->tv = 1.0f;
//	}

//	if (g_gr_bUVFix) {
//		if (ptlv->tu < 0.5f) {
//			ptlv->tu += (0.5f - ptlv->tu) * 0.5f / 256.0f;
//		} else {
//			ptlv->tu += (ptlv->tu - 0.5f) * 1.0f / 256.0f;
//		}
//		ptlv->tu -= (ptlv->tu) * 1.0f / 256.0f;
//		if (ptlv->tv < 0.5f) {
//			ptlv->tv -= (0.5f - ptlv->tv) * 1.0f / 256.0f;
//		} else {
//			ptlv->tv += (ptlv->tv - 0.5f) * 1.0f / 256.0f;
//		}
//	}

}

FX_ENTRY void FX_CALL
grDrawLine( const GrVertex *v1, const GrVertex *v2 )
{
	TRACEFUNCTION("grDrawLine");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			grsubSetTLVertexFromGrVertex(ptlv++, v1);
			grsubSetTLVertexFromGrVertex(ptlv++, v2);
			break;
		}
	case GR_ORIGIN_LOWER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			grsubSetTLVertexFromGrVertex(ptlv++, v2);
			grsubSetTLVertexFromGrVertex(ptlv++, v1);
			break;
		}
	}

	// Draw
	grsubPreRender();
	g_gr_pd3ddevice->DrawPrimitive(D3DPT_LINELIST, D3DFVF_TLVERTEX, tlv, 2, D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);
	grsubPostRender();
}

FX_ENTRY void FX_CALL
grDrawPlanarPolygon( int nverts, const int ilist[], const GrVertex vlist[] )
{
	TRACEFUNCTION("grDrawPlannarPolygon");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	INT i;
	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			GrVertex* pvlist = (GrVertex*)vlist;
			LPINT pilist = (LPINT)ilist;
			LPWORD pidx = idx;
			for (i = 0; i < nverts; i++) {
				grsubSetTLVertexFromGrVertex(ptlv++, pvlist++);
				*pidx++ = *pilist++;
			}
			break;
		}
	case GR_ORIGIN_LOWER_LEFT:
		{
			LPD3DTLVERTEX ptlv = &tlv[nverts];
			GrVertex* pvlist = (GrVertex*)vlist;
			LPINT pilist = (LPINT)ilist;
			LPWORD pidx = idx;
			for (i = 0; i < nverts; i++) {
				grsubSetTLVertexFromGrVertex(--ptlv, pvlist++);
				*pidx++ = *pilist++;
			}
			break;
		}
	}

	// Draw
	grsubPreRender();
	g_gr_pd3ddevice->DrawIndexedPrimitive(GLID3D_POLYGON_D3DPT, D3DFVF_TLVERTEX, tlv, nverts, idx, nverts,
		D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);
	grsubPostRender();
	DEBUG(g_gr_nTrianglePerFrame += nverts / 3);
}

FX_ENTRY void FX_CALL
grDrawPlanarPolygonVertexList( int nverts, const GrVertex vlist[] )
{
	TRACEFUNCTION("grPlanarPolygonVertexList");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	INT i;
	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			GrVertex* pvlist = (GrVertex*)vlist;
			for (i = 0; i < nverts; i++) {
				grsubSetTLVertexFromGrVertex(ptlv++, pvlist++);
			}
			break;
		}
	case GR_ORIGIN_LOWER_LEFT:
		{
			LPD3DTLVERTEX ptlv = &tlv[nverts];
			GrVertex* pvlist = (GrVertex*)vlist;
			for (i = 0; i < nverts; i++) {
				grsubSetTLVertexFromGrVertex(--ptlv, pvlist++);
			}
			break;
		}
	}

	// Draw
	grsubPreRender();
	g_gr_pd3ddevice->DrawPrimitive(GLID3D_POLYGON_D3DPT, D3DFVF_TLVERTEX, tlv, nverts, D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);
	grsubPostRender();
	DEBUG(g_gr_nTrianglePerFrame += nverts / 3);
}

FX_ENTRY void FX_CALL
grDrawPoint( const GrVertex *pt )
{
	TRACEFUNCTION("grDrawPoint");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		grsubSetTLVertexFromGrVertex(tlv, pt);
		break;
	case GR_ORIGIN_LOWER_LEFT:
		grsubSetTLVertexFromGrVertex(tlv, pt);
		break;
	}

	// Draw
	grsubPreRender();
	g_gr_pd3ddevice->DrawPrimitive(D3DPT_POINTLIST, D3DFVF_TLVERTEX, tlv, 1, D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);
	grsubPostRender();
}

FX_ENTRY void FX_CALL
grDrawPolygon( int nverts, const int ilist[], const GrVertex vlist[] )
{
	TRACEFUNCTION("grDrawPolygon");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	INT i;
	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			GrVertex* pvlist = (GrVertex*)vlist;
			LPINT pilist = (LPINT)ilist;
			LPWORD pidx = idx;
			for (i = 0; i < nverts; i++) {
				grsubSetTLVertexFromGrVertex(ptlv++, pvlist++);
				*pidx++ = *pilist++;
			}
			break;
		}
	case GR_ORIGIN_LOWER_LEFT:
		{
			LPD3DTLVERTEX ptlv = &tlv[nverts];
			GrVertex* pvlist = (GrVertex*)vlist;
			LPINT pilist = (LPINT)ilist;
			LPWORD pidx = idx;
			for (i = 0; i < nverts; i++) {
				grsubSetTLVertexFromGrVertex(--ptlv, pvlist++);
				*pidx++ = *pilist++;
			}
			break;
		}
	}

	// Draw
	grsubPreRender();
	g_gr_pd3ddevice->DrawIndexedPrimitive(GLID3D_POLYGON_D3DPT, D3DFVF_TLVERTEX, tlv, nverts, idx, nverts,
		D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);
	grsubPostRender();
	DEBUG(g_gr_nTrianglePerFrame += nverts / 3);
}

FX_ENTRY void FX_CALL
grDrawPolygonVertexList( int nverts, const GrVertex vlist[] )
{
	TRACEFUNCTION("grDrawPolygonVertexList");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	INT i;
	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			GrVertex* pvlist = (GrVertex*)vlist;
			for (i = 0; i < nverts; i++) {
				grsubSetTLVertexFromGrVertex(ptlv++, pvlist++);
			}
			break;
		}
	case GR_ORIGIN_LOWER_LEFT:
		{
			LPD3DTLVERTEX ptlv = &tlv[nverts];
			GrVertex* pvlist = (GrVertex*)vlist;
			for (i = 0; i < nverts; i++) {
				grsubSetTLVertexFromGrVertex(--ptlv, pvlist++);
			}
			break;
		}
	}

	// Draw
	grsubPreRender();
	g_gr_pd3ddevice->DrawPrimitive(GLID3D_POLYGON_D3DPT, D3DFVF_TLVERTEX, tlv, nverts, D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);
	grsubPostRender();
	DEBUG(g_gr_nTrianglePerFrame += nverts / 3);
}

FX_ENTRY void FX_CALL
grDrawTriangle( const GrVertex *a, const GrVertex *b, const GrVertex *c )
{
	TRACEFUNCTION("grDrawTriangle");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			grsubSetTLVertexFromGrVertex(ptlv++, a);
			grsubSetTLVertexFromGrVertex(ptlv++, b);
			grsubSetTLVertexFromGrVertex(ptlv, c);
			break;
		}
	case GR_ORIGIN_LOWER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			grsubSetTLVertexFromGrVertex(ptlv++, c);
			grsubSetTLVertexFromGrVertex(ptlv++, b);
			grsubSetTLVertexFromGrVertex(ptlv, a);
			break;
		}
	}

	// Draw
	grsubPreRender();

#ifndef GLID3D_NOSPECIALEFFECT

	if (!g_gr_bStereo) {

#endif

		g_gr_pd3ddevice->DrawPrimitive(GLID3D_TRIANGLE_D3DPT, D3DFVF_TLVERTEX, tlv, 3, D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);

#ifndef GLID3D_NOSPECIALEFFECT

	} else {
		if (g_gr_state.depth_buffer_mode == GR_DEPTHBUFFER_ZBUFFER) {
			register INT i;
			static FLOAT sx[3], st1[3], st2[3], sttmp;
			sx[0] = tlv[0].sx / 2.0f;
			sx[1] = tlv[1].sx / 2.0f;
			sx[2] = tlv[2].sx / 2.0f;
			for (i = 0; i < 3; i++) {
				if (tlv[i].sz > 0.8f) {
					sttmp = (tlv[i].sz - 0.8f) * (tlv[i].sz - 0.8f) * 5.0f;
					st1[i] = (g_gr_fHarfWidth - sx[i]) * sttmp;
					st2[i] = sx[i] * sttmp;
				} else {
					st1[i] = 0.0f;
					st2[i] = 0.0f;
				}
			}
			tlv[0].sx = sx[0] - st1[0];
			tlv[1].sx = sx[1] - st1[1];
			tlv[2].sx = sx[2] - st1[2];
			g_gr_pd3ddevice->DrawPrimitive(GLID3D_TRIANGLE_D3DPT, D3DFVF_TLVERTEX, tlv, 3, D3DDP_DONOTUPDATEEXTENTS);
			tlv[0].sx = sx[0] + st2[0] + g_gr_fHarfWidth;
			tlv[1].sx = sx[1] + st2[1] + g_gr_fHarfWidth;
			tlv[2].sx = sx[2] + st2[2] + g_gr_fHarfWidth;
			g_gr_pd3ddevice->DrawPrimitive(GLID3D_TRIANGLE_D3DPT, D3DFVF_TLVERTEX, tlv, 3, D3DDP_DONOTUPDATEEXTENTS);
		} else if (g_gr_state.depth_buffer_mode == GR_DEPTHBUFFER_WBUFFER) {
			register INT i;
			static FLOAT sx[3], st1[3], st2[3], sttmp;
			sx[0] = tlv[0].sx / 2.0f;
			sx[1] = tlv[1].sx / 2.0f;
			sx[2] = tlv[2].sx / 2.0f;
			for (i = 0; i < 3; i++) {
				if (tlv[i].rhw > 0.8f) {
					sttmp = (tlv[i].rhw - 0.8f) * (tlv[i].rhw - 0.8f) * 5.0f;
					st1[i] = (g_gr_fHarfWidth - sx[i]) * sttmp;
					st2[i] = sx[i] * sttmp;
				} else {
					st1[i] = 0.0f;
					st2[i] = 0.0f;
				}
			}
			tlv[0].sx = sx[0] - st1[0];
			tlv[1].sx = sx[1] - st1[1];
			tlv[2].sx = sx[2] - st1[2];
			g_gr_pd3ddevice->DrawPrimitive(GLID3D_TRIANGLE_D3DPT, D3DFVF_TLVERTEX, tlv, 3, D3DDP_DONOTUPDATEEXTENTS);
			tlv[0].sx = sx[0] + st2[0] + g_gr_fHarfWidth;
			tlv[1].sx = sx[1] + st2[1] + g_gr_fHarfWidth;
			tlv[2].sx = sx[2] + st2[2] + g_gr_fHarfWidth;
			g_gr_pd3ddevice->DrawPrimitive(GLID3D_TRIANGLE_D3DPT, D3DFVF_TLVERTEX, tlv, 3, D3DDP_DONOTUPDATEEXTENTS);
		}

	}

#endif

	grsubPostRender();
	DEBUG(g_gr_nTrianglePerFrame++);
}

FX_ENTRY void FX_CALL
guAADrawTriangleWithClip( const GrVertex *a, const GrVertex
                         *b, const GrVertex *c)
{
	TRACEFUNCTION("guAADrawTriangleWithClip");

	GLID3D_ANTIALIAS_ENABLE
	guDrawTriangleWithClip(a, b, c);
	GLID3D_ANTIALIAS_DISABLE
}

FX_ENTRY void FX_CALL
guDrawTriangleWithClip(
                       const GrVertex *a,
                       const GrVertex *b,
                       const GrVertex *c
                       )
{
	TRACEFUNCTION("guDrawTriangleWithClip");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			grsubSetTLVertexFromGrVertex(ptlv++, a);
			grsubSetTLVertexFromGrVertex(ptlv++, b);
			grsubSetTLVertexFromGrVertex(ptlv, c);
			break;
		}
	case GR_ORIGIN_LOWER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			grsubSetTLVertexFromGrVertex(ptlv++, c);
			grsubSetTLVertexFromGrVertex(ptlv++, b);
			grsubSetTLVertexFromGrVertex(ptlv, a);
			break;
		}
	}

	// Draw
	grsubPreRender();
	g_gr_pd3ddevice->DrawPrimitive(GLID3D_TRIANGLE_D3DPT, D3DFVF_TLVERTEX, tlv, 3, D3DDP_DONOTUPDATEEXTENTS);
	grsubPostRender();
	DEBUG(g_gr_nTrianglePerFrame++);
}

FX_ENTRY void FX_CALL
guDrawPolygonVertexListWithClip( int nverts, const GrVertex vlist[] )
{
	TRACEFUNCTION("guDrawPolygonVertexListWithClip");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	INT i;
	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			GrVertex* pvlist = (GrVertex*)vlist;
			for (i = 0; i < nverts; i++) {
				grsubSetTLVertexFromGrVertex(ptlv++, pvlist++);
			}
			break;
		}
	case GR_ORIGIN_LOWER_LEFT:
		{
			LPD3DTLVERTEX ptlv = &tlv[nverts];
			GrVertex* pvlist = (GrVertex*)vlist;
			for (i = 0; i < nverts; i++) {
				grsubSetTLVertexFromGrVertex(--ptlv, pvlist++);
			}
			break;
		}
	}

	// Draw
	grsubPreRender();
	g_gr_pd3ddevice->DrawPrimitive(GLID3D_POLYGON_D3DPT, D3DFVF_TLVERTEX, tlv, nverts, D3DDP_DONOTUPDATEEXTENTS);
	grsubPostRender();
	DEBUG(g_gr_nTrianglePerFrame += nverts / 3);
}

/////////////////////////////////////////////////////////////////////////////
