/////////////////////////////////////////////////////////////////////////////
// GliD3D_Config_SpecialEffect.cpp : No comment
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Directive

#include "StdHdr.h"
#pragma hdrstop
#include "GliD3D.h"

/////////////////////////////////////////////////////////////////////////////
// Defines / Variables

/////////////////////////////////////////////////////////////////////////////
// Prototypes

/////////////////////////////////////////////////////////////////////////////
// Implements

FX_ENTRY void FX_CALL
grAlphaBlendFunction(
                     GrAlphaBlendFnc_t rgb_sf,   GrAlphaBlendFnc_t rgb_df,
                     GrAlphaBlendFnc_t alpha_sf, GrAlphaBlendFnc_t alpha_df
                     )
{
	TRACEFUNCTION("grAlphaBlendFunction");

	g_gr_state.alpha_blend_function_rgb_sf = rgb_sf;
	g_gr_state.alpha_blend_function_rgb_df = rgb_df;
	g_gr_state.alpha_blend_function_alpha_sf = alpha_sf;
	g_gr_state.alpha_blend_function_alpha_df = alpha_df;

	if (rgb_sf == GR_BLEND_ONE && rgb_df == GR_BLEND_ZERO) {

		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, FALSE);
		if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_NONE) {
			g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECAL);
		} else {
			g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_MODULATE);
		}

	} else {

		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHABLENDENABLE, TRUE);

		D3DBLEND blend_src, blend_dst;

		switch (rgb_sf) {
		case GR_BLEND_ZERO:
			blend_src = D3DBLEND_ZERO; break;
		case GR_BLEND_ONE:
		case GR_BLEND_PREFOG_COLOR:
			blend_src = D3DBLEND_ONE; break;
		case GR_BLEND_DST_COLOR:
			blend_src = D3DBLEND_DESTCOLOR; break;
		case GR_BLEND_ONE_MINUS_DST_COLOR:
			blend_src = D3DBLEND_INVDESTCOLOR; break;
		case GR_BLEND_SRC_ALPHA:
			blend_src = D3DBLEND_SRCALPHA; break;
		case GR_BLEND_ONE_MINUS_SRC_ALPHA:
			blend_src = D3DBLEND_INVSRCALPHA; break;
		case GR_BLEND_DST_ALPHA:
			blend_src = D3DBLEND_DESTALPHA; break;
		case GR_BLEND_ONE_MINUS_DST_ALPHA:
			blend_src = D3DBLEND_INVDESTALPHA; break;
		}

		switch (rgb_df) {
		case GR_BLEND_ZERO:
			blend_dst = D3DBLEND_ZERO; break;
		case GR_BLEND_ONE:
		case GR_BLEND_PREFOG_COLOR:
			blend_dst = D3DBLEND_ONE; break;
		case GR_BLEND_SRC_COLOR:
			blend_dst = D3DBLEND_SRCCOLOR; break;
		case GR_BLEND_ONE_MINUS_SRC_COLOR:
			blend_dst = D3DBLEND_INVSRCCOLOR; break;
		case GR_BLEND_SRC_ALPHA:
			blend_dst = D3DBLEND_SRCALPHA; break;
		case GR_BLEND_ONE_MINUS_SRC_ALPHA:
			blend_dst = D3DBLEND_INVSRCALPHA; break;
		case GR_BLEND_DST_ALPHA:
			blend_dst = D3DBLEND_DESTALPHA; break;
		case GR_BLEND_ONE_MINUS_DST_ALPHA:
			blend_dst = D3DBLEND_INVDESTALPHA; break;
		}

		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_SRCBLEND,  blend_src);
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_DESTBLEND, blend_dst);

		if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_NONE) {
			if (g_gr_state.alpha_combine_local == GR_COMBINE_LOCAL_NONE) {
				g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECAL);
			} else {
				g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECALALPHA);
			}
		} else {
			if (g_gr_state.alpha_combine_local == GR_COMBINE_LOCAL_NONE) {
				g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_MODULATE);
			} else {
				g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_MODULATEALPHA);
			}
		}

	}
}

FX_ENTRY void FX_CALL
grAlphaCombine(
               GrCombineFunction_t function, GrCombineFactor_t factor,
               GrCombineLocal_t local, GrCombineOther_t other,
               FxBool invert
               )
{
	TRACEFUNCTION("grAlphaCombine");

	g_gr_state.alpha_combine_function = function;
	g_gr_state.alpha_combine_factor = factor;
	g_gr_state.alpha_combine_local = local;
	g_gr_state.alpha_combine_other = other;
}

FX_ENTRY void FX_CALL
grAlphaControlsITRGBLighting( FxBool enable )
{
	TRACEFUNCTION("grAlphaControlsITRGBLighting");

}

FX_ENTRY void FX_CALL
grAlphaTestFunction( GrCmpFnc_t function )
{
	TRACEFUNCTION("grAlphaTestFunction");

	g_gr_state.alpha_test_function = function;
	switch (function) {
	case GR_CMP_NEVER    :
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHAFUNC, D3DCMP_NEVER);
		break;
	case GR_CMP_LESS     :
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHAFUNC, D3DCMP_LESS);
		break;
	case GR_CMP_EQUAL    :
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHAFUNC, D3DCMP_EQUAL);
		break;
	case GR_CMP_LEQUAL   :
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHAFUNC, D3DCMP_LESSEQUAL);
		break;
	case GR_CMP_GREATER  :
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHAFUNC, D3DCMP_GREATER);
		break;
	case GR_CMP_NOTEQUAL :
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHAFUNC, D3DCMP_NOTEQUAL);
		break;
	case GR_CMP_GEQUAL   :
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, TRUE);
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHAFUNC, D3DCMP_GREATEREQUAL);
		break;
	case GR_CMP_ALWAYS   :
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHATESTENABLE, FALSE);
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHAFUNC, D3DCMP_ALWAYS);
		break;
	}
}

FX_ENTRY void FX_CALL
grAlphaTestReferenceValue( GrAlpha_t value )
{
	TRACEFUNCTION("grAlphaTestReferenceValue");

	g_gr_state.alpha_test_reference_value = value;
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ALPHAREF, value);
}

FX_ENTRY void FX_CALL 
grChromakeyMode( GrChromakeyMode_t mode )
{
	TRACEFUNCTION("grChromakeyMode");

	g_gr_state.chromakey_mode = mode;
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_COLORKEYENABLE, mode == GR_CHROMAKEY_ENABLE ? TRUE : FALSE);
}

FX_ENTRY void FX_CALL 
grChromakeyValue( GrColor_t value )
{
	TRACEFUNCTION("grChromakeyValue");

	D3DCOLOR cr = grsubConvColor(value);
	g_gr_state.chromakey_value = cr;
}

FX_ENTRY void FX_CALL 
grClipWindow( FxU32 minx, FxU32 miny, FxU32 maxx, FxU32 maxy )
{
	TRACEFUNCTION("grClipWindow");

	g_gr_state.clip_window_minx = minx;
	g_gr_state.clip_window_miny = miny;
	g_gr_state.clip_window_maxx = maxx;
	g_gr_state.clip_window_maxy = maxy;

	FLOAT fLeft, fTop, fRight, fBottom;

	if (g_gr_bScaleFix) {
		fLeft   = -1.0f + (FLOAT)minx * 2.0f / (FLOAT)g_gr_nWidth * g_gr_fWidthScale;
		fTop    = -1.0f + (FLOAT)miny * 2.0f / (FLOAT)g_gr_nHeight * g_gr_fHeightScale;
		fRight  = -1.0f + (FLOAT)maxx * 2.0f / (FLOAT)g_gr_nWidth * g_gr_fWidthScale;
		fBottom = -1.0f + (FLOAT)maxy * 2.0f / (FLOAT)g_gr_nHeight * g_gr_fHeightScale;
	} else {
		fLeft   = -1.0f + (FLOAT)minx * 2.0f / (FLOAT)g_gr_nWidth;
		fTop    = -1.0f + (FLOAT)miny * 2.0f / (FLOAT)g_gr_nHeight;
		fRight  = -1.0f + (FLOAT)maxx * 2.0f / (FLOAT)g_gr_nWidth;
		fBottom = -1.0f + (FLOAT)maxy * 2.0f / (FLOAT)g_gr_nHeight;
	}
	if (fLeft < -1.0f) fLeft = -1.0f;
	if (fTop < -1.0f) fTop = -1.0f;
	if (fRight > 1.0f) fRight = 1.0f;
	if (fBottom > 1.0f) fBottom = 1.0f;
	GB3DSetViewport(GB3DVIEWPORT_DEFAULT, fLeft, fTop, fRight, fBottom);
}

FX_ENTRY void FX_CALL 
grColorCombine(
               GrCombineFunction_t function, GrCombineFactor_t factor,
               GrCombineLocal_t local, GrCombineOther_t other,
               FxBool invert )
{
	TRACEFUNCTION("grColorCombine");

	g_gr_state.color_combine_function = function;
	g_gr_state.color_combine_factor = factor;
	g_gr_state.color_combine_local = local;
	g_gr_state.color_combine_other = other;

	if (g_gr_state.color_combine_function == GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA)
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_SPECULARENABLE, TRUE);
	else
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_SPECULARENABLE, FALSE);

	if (g_gr_state.color_combine_factor == GR_COMBINE_FACTOR_ONE) {
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_DECAL);
	} else {
		g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_TEXTUREMAPBLEND, D3DTBLEND_MODULATE);
	}

	g_gr_tmu[0].texture_source_start_address_old = -1;

}

FX_ENTRY void FX_CALL
grColorMask( FxBool rgb, FxBool a )
{
	TRACEFUNCTION("grColorMask");

	g_gr_state.color_mask_rgb = rgb;
	g_gr_state.color_mask_a = a;

	DWORD dwMask = 0;
	if (rgb == FXTRUE) dwMask |= 0x00ffffff;
	if (a == FXTRUE) dwMask |= 0xff000000;
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_PLANEMASK, dwMask);
}

FX_ENTRY void FX_CALL 
grCullMode( GrCullMode_t mode )
{
	TRACEFUNCTION("grCullMode");

	g_gr_state.cull_mode = mode;
	switch (mode) {
	case GR_CULL_DISABLE : g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_NONE); break;
	case GR_CULL_NEGATIVE: g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_CCW); break;
	case GR_CULL_POSITIVE: g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_CULLMODE, D3DCULL_CW); break;
	}
}

FX_ENTRY void FX_CALL 
grConstantColorValue( GrColor_t value )
{
	TRACEFUNCTION("grConstantColorValue");

	D3DCOLOR cr = grsubConvColor(value);
	g_gr_state.constant_color_value = cr;
	g_gr_state.constant_color_value_color = cr & 0x00ffffff;
	g_gr_state.constant_color_value_alpha = cr & 0xff000000;
}

FX_ENTRY void FX_CALL 
grDepthBiasLevel( FxI32 level )
{
	TRACEFUNCTION("grDepthBiasLevel");

	g_gr_state.depth_bias_level = level;
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ZBIAS, g_gr_state.depth_bias_level >> 12);
}

FX_ENTRY void FX_CALL 
grDepthBufferFunction( GrCmpFnc_t function )
{
	TRACEFUNCTION("grDepthBufferFunction");

	g_gr_state.depth_buffer_function = function;
	switch (function) {
	case GR_CMP_NEVER    : g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_NEVER); break;
	case GR_CMP_LESS     : g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_LESS); break;
	case GR_CMP_EQUAL    : g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_EQUAL); break;
	case GR_CMP_LEQUAL   : g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_LESSEQUAL); break;
	case GR_CMP_GREATER  : g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_GREATER); break;
	case GR_CMP_NOTEQUAL : g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_NOTEQUAL); break;
	case GR_CMP_GEQUAL   : g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_GREATEREQUAL); break;
	case GR_CMP_ALWAYS   : g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ZFUNC, D3DCMP_ALWAYS); break;
	}
}

FX_ENTRY void FX_CALL 
grDepthBufferMode( GrDepthBufferMode_t mode )
{
	TRACEFUNCTION("grDepthBufferMode");

	g_gr_state.depth_buffer_mode = mode;
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ZENABLE,
		(mode == GR_DEPTHBUFFER_DISABLE) ? D3DZB_FALSE :
		(mode == GR_DEPTHBUFFER_ZBUFFER) ? D3DZB_TRUE :
		(g_gr_bUseZForW) ? D3DZB_TRUE : D3DZB_USEW);
	TRACE("GliD3D info : Depth buffer mode = %s\n",
		mode == GR_DEPTHBUFFER_DISABLE ? "DISABLE" :
		mode == GR_DEPTHBUFFER_ZBUFFER ? "ZBUFFER" : "WBUFFER"
	);
}

FX_ENTRY void FX_CALL 
grDepthMask( FxBool mask )
{
	TRACEFUNCTION("grDepthMask");

	g_gr_state.depth_mask = mask;
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ZWRITEENABLE, (mask == FXTRUE));
}

FX_ENTRY void FX_CALL 
grDisableAllEffects( void )
{
	TRACEFUNCTION("grDisableAllEffects");

	grAlphaBlendFunction(GR_BLEND_ONE, GR_BLEND_ZERO, GR_BLEND_ONE, GR_BLEND_ZERO);
	grAlphaTestFunction(GR_CMP_ALWAYS);
	grChromakeyMode(GR_CHROMAKEY_DISABLE);
	grDepthBufferMode(GR_DEPTHBUFFER_DISABLE);
	grFogMode(GR_FOG_DISABLE);
}

FX_ENTRY void FX_CALL 
grDitherMode( GrDitherMode_t mode )
{
	TRACEFUNCTION("grDitherMode");

	g_gr_state.dither_mode = mode;
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_DITHERENABLE, (mode != GR_DITHER_DISABLE));
}

FX_ENTRY void FX_CALL 
grFogColorValue( GrColor_t fogcolor )
{
	TRACEFUNCTION("grFogColorValue");

	g_gr_state.fog_color = fogcolor;
	D3DCOLOR cr = grsubConvColor(fogcolor);
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGCOLOR, cr);
	g_gr_state.fog_d3dcolor = cr;
}

FX_ENTRY void FX_CALL 
grFogMode( GrFogMode_t mode )
{
	TRACEFUNCTION("grFogMode");

	if (g_gr_bNoFog) return;

	g_gr_state.fog_mode = mode;
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGENABLE, (mode != GR_FOG_DISABLE));

	if (g_gr_state.fog_mode & GR_FOG_WITH_TABLE_ON_Q) {
		INT nFogTableValidate = (INT)g_gr_fogtable[GLID3D_FOGTABLE_VALIDATE];
		if (nFogTableValidate == GLID3D_FOGTABLE_VALIDATE_VALUE) {
			INT nFogMode = (INT)g_gr_fogtable[GLID3D_FOGTABLE_MODE];
			switch (nFogMode) {
			case GLID3D_FOGMODE_EXP:
				{
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEDENSITY, *((DWORD *)(&g_gr_fogtable[GLID3D_FOGTABLE_DENSITY])));
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEMODE, D3DFOG_EXP);
					break;
				}
			case GLID3D_FOGMODE_EXP2:
				{
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEDENSITY, *((DWORD *)(&g_gr_fogtable[GLID3D_FOGTABLE_DENSITY])));
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEMODE, D3DFOG_EXP2);
					break;
				}
			case GLID3D_FOGMODE_LINEAR:
				{
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLESTART, *((DWORD *)(&g_gr_fogtable[GLID3D_FOGTABLE_NEARZ])));
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEEND, *((DWORD *)(&g_gr_fogtable[GLID3D_FOGTABLE_FARZ])));
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEMODE, D3DFOG_LINEAR);
					break;
				}
			}
		} else {
//			TRACE("Invalid fogtable it was generated by non glide function\n");
		}
	}
}

FX_ENTRY void FX_CALL 
grFogTable( const GrFog_t ft[] )
{
	TRACEFUNCTION("grFogTable");

	if (g_gr_bNoFog) return;

	CopyMemory(g_gr_fogtable, ft, sizeof(BYTE) * 64);

	if (g_gr_state.fog_mode & GR_FOG_WITH_TABLE_ON_Q) {
		INT nFogTableValidate = (INT)g_gr_fogtable[GLID3D_FOGTABLE_VALIDATE];
		if (nFogTableValidate == GLID3D_FOGTABLE_VALIDATE_VALUE) {
			INT nFogMode = (INT)g_gr_fogtable[GLID3D_FOGTABLE_MODE];
			switch (nFogMode) {
			case GLID3D_FOGMODE_EXP:
				{
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEDENSITY, *((DWORD *)(&g_gr_fogtable[GLID3D_FOGTABLE_DENSITY])));
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEMODE, D3DFOG_EXP);
					break;
				}
			case GLID3D_FOGMODE_EXP2:
				{
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEDENSITY, *((DWORD *)(&g_gr_fogtable[GLID3D_FOGTABLE_DENSITY])));
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEMODE, D3DFOG_EXP2);
					break;
				}
			case GLID3D_FOGMODE_LINEAR:
				{
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLESTART, *((DWORD *)(&g_gr_fogtable[GLID3D_FOGTABLE_NEARZ])));
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEEND, *((DWORD *)(&g_gr_fogtable[GLID3D_FOGTABLE_FARZ])));
					g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_FOGTABLEMODE, D3DFOG_LINEAR);
					break;
				}
			}
		} else {
//			TRACE("Invalid fogtable it was generated by non glide function\n");
		}
	}
}

FX_ENTRY void FX_CALL 
grLoadGammaTable( FxU32 nentries, FxU32 *red, FxU32 *green, FxU32 *blue)
{
	TRACEFUNCTION("grLoadGammaTable");

}

FX_ENTRY void FX_CALL
grSplash(float x, float y, float width, float height, FxU32 frame)
{
	TRACEFUNCTION("grSplash");

	if (frame == 0) {
/*
		//GrState state;
		BYTE state[512];
		GB3DEndRender();
		grGlideGetState(&state);
		GB3DSplashLetSoftware();
		GB3DBeginRender();
		grGlideSetState(&state);
*/
	}
}

FX_ENTRY FxU32 FX_CALL 
grGet( FxU32 pname, FxU32 plength, FxI32 *params )
{
	TRACEFUNCTION("grGet");

	switch (pname) {
	case GR_WDEPTH_MIN_MAX           : params[0] = 0; params[1] = 0xffff; TRACE("GliD3D info : WDEPTH\n"); break;
	case GR_ZDEPTH_MIN_MAX           : params[0] = 0xffff; params[1] = 0; TRACE("GliD3D info : ZDEPTH\n"); break;
	case GR_GLIDE_STATE_SIZE         : params[0] = sizeof(GLID3D_STATE); TRACE("GliD3D info : Glide State size\n"); break;
	case GR_GLIDE_VERTEXLAYOUT_SIZE  : params[0] = sizeof(GLID3D_VERTEXLAYOUT); TRACE("GliD3D info : Vertex Layout size\n"); break;
	}
	return plength;
}

FX_ENTRY const char * FX_CALL 
grGetString( FxU32 pname )
{
	TRACEFUNCTION("grGetString");

	switch (pname) {
	case GR_EXTENSION:
		TRACE("GliD3D info : Extension\n");
		return "";
	case GR_HARDWARE:
		TRACE("GliD3D info : Hardware\n");
		return "Voodoo Graphics";
	case GR_RENDERER:
		TRACE("GliD3D info : Renderer\n");
		return "Glide";
	case GR_VENDOR:
		TRACE("GliD3D info : Vendor\n");
		return "Let software(TM)";
	case GR_VERSION:
		TRACE("GliD3D info : Version\n");
		return "3.0";
	}
	return "";
}

FX_ENTRY FxI32 FX_CALL 
grQueryResolutions( const GrResolution *resTemplate, GrResolution *output )
{
	TRACEFUNCTION("grQueryResolutions");

	DWORD dwSize = 0;
	if (resTemplate->resolution == GR_QUERY_ANY) {
		dwSize = sizeof(GrResolution) * 5;
		if (output) {
			GrResolution* op = output;
			op->resolution = GR_RESOLUTION_512x384;
			op->refresh = GR_REFRESH_60Hz;
			op->numColorBuffers = 3;
			op->numAuxBuffers = 1;
			op++;
			op->resolution = GR_RESOLUTION_640x400;
			op->refresh = GR_REFRESH_60Hz;
			op->numColorBuffers = 3;
			op->numAuxBuffers = 1;
			op++;
			op->resolution = GR_RESOLUTION_640x480;
			op->refresh = GR_REFRESH_60Hz;
			op->numColorBuffers = 3;
			op->numAuxBuffers = 1;
			op++;
			op->resolution = GR_RESOLUTION_800x600;
			op->refresh = GR_REFRESH_60Hz;
			op->numColorBuffers = 3;
			op->numAuxBuffers = 1;
			op++;
			op->resolution = GR_RESOLUTION_1024x768;
			op->refresh = GR_REFRESH_60Hz;
			op->numColorBuffers = 2;
			op->numAuxBuffers = 1;
			op++;
		}
	} else {
		dwSize = sizeof(GrResolution);
		if (output) {
			GrResolution* op = output;
			op->resolution = resTemplate->resolution;
			op->refresh = GR_REFRESH_60Hz;
			op->numColorBuffers = 2;
			op->numAuxBuffers = 1;
		}
	}

	return dwSize;
}

FX_ENTRY FxBool FX_CALL 
grReset( FxU32 what )
{
	TRACEFUNCTION("grReset");

	return 1;
}

FX_ENTRY GrProc FX_CALL
grGetProcAddress( char *procName )
{
	TRACEFUNCTION("grGetProcAddress");

	return NULL;
}

FX_ENTRY void FX_CALL 
grEnable( GrEnableMode_t mode )
{
	TRACEFUNCTION("grEnable");
}

FX_ENTRY void FX_CALL 
grDisable( GrEnableMode_t mode )
{
	TRACEFUNCTION("grDisable");
}

FX_ENTRY void FX_CALL 
grCoordinateSpace( GrCoordinateSpaceMode_t mode )
{
	TRACEFUNCTION("grCoorinateSpace");

//	g_gr_state.coorinatespacemode = mode;
}

FX_ENTRY void FX_CALL 
grDepthRange( FxFloat n, FxFloat f )
{
	TRACEFUNCTION("grDepthRange");
}

FX_ENTRY void FX_CALL 
grViewport( FxI32 x, FxI32 y, FxI32 width, FxI32 height )
{
	TRACEFUNCTION("grViewport");

	grClipWindow(x, y, x + width, y + height);
}

/////////////////////////////////////////////////////////////////////////////
