/////////////////////////////////////////////////////////////////////////////
// glid3d_render.cpp : No comment
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
// Directive

#include "stdhdr.h"
#pragma hdrstop
#include "glid3d.h"

/////////////////////////////////////////////////////////////////////////////
// Defines / Variables

#define GLID3D_ANTIALIAS_ENABLE \
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ANTIALIAS, D3DANTIALIAS_SORTINDEPENDENT); \
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_EDGEANTIALIAS, TRUE);

#define GLID3D_ANTIALIAS_DISABLE \
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_ANTIALIAS, D3DANTIALIAS_NONE); \
	g_gr_pd3ddevice->SetRenderState(D3DRENDERSTATE_EDGEANTIALIAS, FALSE);

//#define GLID3D_D3DDP_DONOTCLIP	(D3DDP_DONOTCLIP)
#define GLID3D_D3DDP_DONOTCLIP	(0)

static D3DTLVERTEX tlv[4096];
static WORD idx[4096];
static FLOAT f256 = 256.0f;
static FLOAT f4096 = 4096.0f;
static FLOAT f65535 = 65535.0f;
static FLOAT f65536 = 65536.0f;

/////////////////////////////////////////////////////////////////////////////
// Prototypes

/////////////////////////////////////////////////////////////////////////////
// Implements

inline VOID
grsubSetTLVertexFromGrVertex(LPD3DTLVERTEX ptlv, const void* pv)
{
	FLOAT pvx = *(FLOAT*)((BYTE*)pv + g_gr_vertex_layout.nXY + 0);
	FLOAT pvy = *(FLOAT*)((BYTE*)pv + g_gr_vertex_layout.nXY + 4);
	FLOAT pvr = *(FLOAT*)((BYTE*)pv + g_gr_vertex_layout.nRGB + 0);
	FLOAT pvg = *(FLOAT*)((BYTE*)pv + g_gr_vertex_layout.nRGB + 4);
	FLOAT pvb = *(FLOAT*)((BYTE*)pv + g_gr_vertex_layout.nRGB + 8);
	FLOAT pva = *(FLOAT*)((BYTE*)pv + g_gr_vertex_layout.nA   + 0); 
	FLOAT pvooz = *(FLOAT*)((BYTE*)pv + g_gr_vertex_layout.nZ + 0);
	FLOAT pvoow = *(FLOAT*)((BYTE*)pv + g_gr_vertex_layout.nQ + 0);
	FLOAT pv0sow = *(FLOAT*)((BYTE*)pv + g_gr_vertex_layout.nST0 + 0);
	FLOAT pv0tow = *(FLOAT*)((BYTE*)pv + g_gr_vertex_layout.nST0 + 4);

#ifndef GLID3D_NOSCALINGFIX

	if (g_gr_bScaleFix) {
		if (pvx < f4096)
			ptlv->sx = pvx * g_gr_fWidthScale;
		else
			ptlv->sx = (pvx - (FLOAT)(((DWORD)pvx) & 0xfffff000)) * g_gr_fWidthScale;

		if (g_gr_state.origin_location == GR_ORIGIN_UPPER_LEFT) {
			if (pvy < f4096)
				ptlv->sy = pvy * g_gr_fHeightScale;
			else
				ptlv->sy = (pvy - (FLOAT)(((DWORD)pvy) & 0xfffff000)) * g_gr_fHeightScale;

		} else {
			if (pvy < f4096)
				ptlv->sy = ((FLOAT)g_gr_nHeight - pvy) * g_gr_fHeightScale;
			else
				ptlv->sy = ((FLOAT)g_gr_nHeight - (FLOAT)(((DWORD)pvy) & 0xfffff000)) * g_gr_fHeightScale;

		}
	} else {

#endif

		if (pvx < f4096)
			ptlv->sx = pvx;
		else
			ptlv->sx = pvx - (FLOAT)(((DWORD)pvx) & 0xfffff000);

		if (g_gr_state.origin_location == GR_ORIGIN_UPPER_LEFT) {
			if (pvy < f4096)
				ptlv->sy = pvy;
			else
				ptlv->sy = pvy - (FLOAT)(((DWORD)pvy) & 0xfffff000);

		} else {
			if (pvy < f4096)
				ptlv->sy = (FLOAT)g_gr_nHeight - pvy;
			else
				ptlv->sy = (FLOAT)g_gr_nHeight - (FLOAT)(((DWORD)pvy) & 0xfffff000);

		}

#ifndef GLID3D_NOSCALINGFIX

	}

#endif

	switch (g_gr_state.depth_buffer_mode) {
	case GR_DEPTHBUFFER_ZBUFFER:
		ptlv->sz = f65535 / pvooz;
		ptlv->rhw = pvooz / f65535;
		break;
	case GR_DEPTHBUFFER_WBUFFER:
		if (!g_gr_bUseZForW) {
			ptlv->sz = 1.0f - pvoow;
			ptlv->rhw = pvoow;
		} else {
			ptlv->sz = 1.0f / pvoow / f65535;
			ptlv->rhw = 1.0f / ptlv->sz;
		}
		break;
	case GR_DEPTHBUFFER_DISABLE:
		ptlv->sz = f65535 / pvooz;
		ptlv->rhw = pvooz / f65535;
	}

//	if (g_gr_bStereo) {
//
//		if (g_gr_state.depth_buffer_mode == GR_DEPTHBUFFER_ZBUFFER) {
//			ptlv->sx += (10.0f * ptlv->sz * ptlv->sz * ptlv->sz) * (g_gr_bLeft ? 1.0f : -1.0f);
//		} else if (g_gr_state.depth_buffer_mode == GR_DEPTHBUFFER_WBUFFER) {
//			ptlv->sx += (10.0f * (1.0f - ptlv->rhw * ptlv->rhw * ptlv->rhw)) * (g_gr_bLeft ? 1.0f : -1.0f);
//		}
//
//	}	
	
	if (g_gr_state.alpha_blend_function_rgb_sf != GR_BLEND_PREFOG_COLOR) {

		switch (g_gr_state.color_combine_local) {
		case GR_COMBINE_LOCAL_ITERATED:
			switch (g_gr_state.alpha_combine_local) {
			case GR_COMBINE_LOCAL_ITERATED:
//				ptlv->color = ((DWORD)pvr << 16) | ((DWORD)pvg <<  8) | ((DWORD)pvb <<  0) | ((DWORD)pva << 24);
				ptlv->color = ((((DWORD)pva << 8) | ((DWORD)pvr)) << 16) | ((DWORD)pvg <<  8) | ((DWORD)pvb);
//				ptlv->color = (DWORD)(pva * 256.0f * 65536.0f + pvr * 65536.0f + pvg * 256.0f + pvb);
				break;
			case GR_COMBINE_LOCAL_CONSTANT:
				ptlv->color = ((DWORD)pvr << 16) | ((DWORD)pvg <<  8) | ((DWORD)pvb <<  0) | (g_gr_state.constant_color_value_alpha);
//				ptlv->color = (DWORD)(pvr * 65536.0f + pvg * 256.0f + pvb) + g_gr_state.constant_color_value_alpha;
				break;
			}
			break;
		case GR_COMBINE_LOCAL_CONSTANT:
			switch (g_gr_state.alpha_combine_local) {
			case GR_COMBINE_LOCAL_ITERATED:
				ptlv->color = g_gr_state.constant_color_value_color | ((DWORD)pva << 24);
				break;
			case GR_COMBINE_LOCAL_CONSTANT:
				ptlv->color = g_gr_state.constant_color_value;
				break;
			}
			break;
		}
	} else {
//		ptlv->color = g_gr_state.fog_d3dcolor;
	}
	switch (g_gr_state.color_combine_function) {
	case GR_COMBINE_FUNCTION_SCALE_OTHER_ADD_LOCAL_ALPHA:
//		FLOAT fF;
//		fF = pv->a + pv->a * 255.0f;
//		fF = fF + fF * 65536.0f;
//		ptlv->specular = (DWORD)fF;
		ptlv->specular = ((DWORD)pva) | ((DWORD)pva << 8);
		ptlv->specular |= ptlv->specular << 16;
		break;
	default:
		ptlv->specular = 0;
	}
	ptlv->tu = ((pv0sow / pvoow) + 0.2f) / f256;
	ptlv->tv = ((pv0tow / pvoow) + 0.2f) / f256;
//	if (g_gr_tmu[0].s_clampmode == GR_TEXTURECLAMP_CLAMP) {
//		if (ptlv->tu < 0.0f) ptlv->tu = 0.0f;
//		if (ptlv->tu > 1.0f) ptlv->tu = 1.0f;
//	}
//	if (g_gr_tmu[0].t_clampmode == GR_TEXTURECLAMP_CLAMP) {
//		if (ptlv->tv < 0.0f) ptlv->tv = 0.0f;
//		if (ptlv->tv > 1.0f) ptlv->tv = 1.0f;
//	}

//	if (g_gr_bUVFix) {
//		if (ptlv->tu < 0.5f) {
//			ptlv->tu += (0.5f - ptlv->tu) * 0.5f / 256.0f;
//		} else {
//			ptlv->tu += (ptlv->tu - 0.5f) * 1.0f / 256.0f;
//		}
//		ptlv->tu -= (ptlv->tu) * 1.0f / 256.0f;
//		if (ptlv->tv < 0.5f) {
//			ptlv->tv -= (0.5f - ptlv->tv) * 1.0f / 256.0f;
//		} else {
//			ptlv->tv += (ptlv->tv - 0.5f) * 1.0f / 256.0f;
//		}
//	}

}

FX_ENTRY void FX_CALL
grDrawPoint( const void *pt )
{
	TRACEFUNCTION("grDrawPoint");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		grsubSetTLVertexFromGrVertex(tlv, pt);
		break;
	case GR_ORIGIN_LOWER_LEFT:
		grsubSetTLVertexFromGrVertex(tlv, pt);
		break;
	}

	// Draw
	grsubPreRender();
	g_gr_pd3ddevice->DrawPrimitive(D3DPT_POINTLIST, D3DFVF_TLVERTEX, tlv, 1, D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);
	grsubPostRender();
}

FX_ENTRY void FX_CALL
grDrawLine( const void *v1, const void *v2 )
{
	TRACEFUNCTION("grDrawLine");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			grsubSetTLVertexFromGrVertex(ptlv++, v1);
			grsubSetTLVertexFromGrVertex(ptlv++, v2);
			break;
		}
	case GR_ORIGIN_LOWER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			grsubSetTLVertexFromGrVertex(ptlv++, v2);
			grsubSetTLVertexFromGrVertex(ptlv++, v1);
			break;
		}
	}

	// Draw
	grsubPreRender();
	g_gr_pd3ddevice->DrawPrimitive(D3DPT_LINELIST, D3DFVF_TLVERTEX, tlv, 2, D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);
	grsubPostRender();
}

FX_ENTRY void FX_CALL
grDrawTriangle( const void *a, const void *b, const void *c )
{
	TRACEFUNCTION("grDrawTriangle");

//	if (g_gr_state.color_combine_local == GR_COMBINE_LOCAL_CONSTANT &&
//		g_gr_state.chromakey_mode == GR_CHROMAKEY_ENABLE &&
//		g_gr_state.chromakey_value == g_gr_state.constant_color_value
//	) return;

	switch (g_gr_state.origin_location) {
	case GR_ORIGIN_UPPER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			grsubSetTLVertexFromGrVertex(ptlv++, a);
			grsubSetTLVertexFromGrVertex(ptlv++, b);
			grsubSetTLVertexFromGrVertex(ptlv, c);
			break;
		}
	case GR_ORIGIN_LOWER_LEFT:
		{
			LPD3DTLVERTEX ptlv = tlv;
			grsubSetTLVertexFromGrVertex(ptlv++, c);
			grsubSetTLVertexFromGrVertex(ptlv++, b);
			grsubSetTLVertexFromGrVertex(ptlv, a);
			break;
		}
	}

	// Draw
	grsubPreRender();

#ifndef GLID3D_NOSPECIALEFFECT

	if (!g_gr_bStereo) {

#endif

		g_gr_pd3ddevice->DrawPrimitive(GLID3D_TRIANGLE_D3DPT, D3DFVF_TLVERTEX, tlv, 3, D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);

#ifndef GLID3D_NOSPECIALEFFECT

	} else {
		if (g_gr_state.depth_buffer_mode == GR_DEPTHBUFFER_ZBUFFER) {
			register INT i;
			static FLOAT sx[3], st1[3], st2[3], sttmp;
			sx[0] = tlv[0].sx / 2.0f;
			sx[1] = tlv[1].sx / 2.0f;
			sx[2] = tlv[2].sx / 2.0f;
			for (i = 0; i < 3; i++) {
				if (tlv[i].sz > 0.8f) {
					sttmp = (tlv[i].sz - 0.8f) * (tlv[i].sz - 0.8f) * 5.0f;
					st1[i] = (g_gr_fHarfWidth - sx[i]) * sttmp;
					st2[i] = sx[i] * sttmp;
				} else {
					st1[i] = 0.0f;
					st2[i] = 0.0f;
				}
			}
			tlv[0].sx = sx[0] - st1[0];
			tlv[1].sx = sx[1] - st1[1];
			tlv[2].sx = sx[2] - st1[2];
			g_gr_pd3ddevice->DrawPrimitive(GLID3D_TRIANGLE_D3DPT, D3DFVF_TLVERTEX, tlv, 3, D3DDP_DONOTUPDATEEXTENTS);
			tlv[0].sx = sx[0] + st2[0] + g_gr_fHarfWidth;
			tlv[1].sx = sx[1] + st2[1] + g_gr_fHarfWidth;
			tlv[2].sx = sx[2] + st2[2] + g_gr_fHarfWidth;
			g_gr_pd3ddevice->DrawPrimitive(GLID3D_TRIANGLE_D3DPT, D3DFVF_TLVERTEX, tlv, 3, D3DDP_DONOTUPDATEEXTENTS);
		} else if (g_gr_state.depth_buffer_mode == GR_DEPTHBUFFER_WBUFFER) {
			register INT i;
			static FLOAT sx[3], st1[3], st2[3], sttmp;
			sx[0] = tlv[0].sx / 2.0f;
			sx[1] = tlv[1].sx / 2.0f;
			sx[2] = tlv[2].sx / 2.0f;
			for (i = 0; i < 3; i++) {
				if (tlv[i].rhw > 0.8f) {
					sttmp = (tlv[i].rhw - 0.8f) * (tlv[i].rhw - 0.8f) * 5.0f;
					st1[i] = (g_gr_fHarfWidth - sx[i]) * sttmp;
					st2[i] = sx[i] * sttmp;
				} else {
					st1[i] = 0.0f;
					st2[i] = 0.0f;
				}
			}
			tlv[0].sx = sx[0] - st1[0];
			tlv[1].sx = sx[1] - st1[1];
			tlv[2].sx = sx[2] - st1[2];
			g_gr_pd3ddevice->DrawPrimitive(GLID3D_TRIANGLE_D3DPT, D3DFVF_TLVERTEX, tlv, 3, D3DDP_DONOTUPDATEEXTENTS);
			tlv[0].sx = sx[0] + st2[0] + g_gr_fHarfWidth;
			tlv[1].sx = sx[1] + st2[1] + g_gr_fHarfWidth;
			tlv[2].sx = sx[2] + st2[2] + g_gr_fHarfWidth;
			g_gr_pd3ddevice->DrawPrimitive(GLID3D_TRIANGLE_D3DPT, D3DFVF_TLVERTEX, tlv, 3, D3DDP_DONOTUPDATEEXTENTS);
		}

	}

#endif

	grsubPostRender();
	DEBUG(g_gr_nTrianglePerFrame++);
}

FX_ENTRY void FX_CALL
grVertexLayout(FxU32 param, FxI32 offset, FxU32 mode)
{
	switch (param) {
	case GR_PARAM_XY  : g_gr_vertex_layout.nXY  = (mode == GR_PARAM_ENABLE ? offset : 0); break;
	case GR_PARAM_RGB : g_gr_vertex_layout.nRGB = (mode == GR_PARAM_ENABLE ? offset : 0); break;
	case GR_PARAM_A   : g_gr_vertex_layout.nA   = (mode == GR_PARAM_ENABLE ? offset : 0); break;
	case GR_PARAM_Z   : g_gr_vertex_layout.nZ   = (mode == GR_PARAM_ENABLE ? offset : 0); break;
	case GR_PARAM_Q   : g_gr_vertex_layout.nQ   = (mode == GR_PARAM_ENABLE ? offset : 0); break;
	case GR_PARAM_ST0 : g_gr_vertex_layout.nST0 = (mode == GR_PARAM_ENABLE ? offset : 0); break;
	case GR_PARAM_Q0  : g_gr_vertex_layout.nQ0  = (mode == GR_PARAM_ENABLE ? offset : 0); break;
	case GR_PARAM_ST1 : g_gr_vertex_layout.nST1 = (mode == GR_PARAM_ENABLE ? offset : 0); break;
	case GR_PARAM_Q1  : g_gr_vertex_layout.nQ1  = (mode == GR_PARAM_ENABLE ? offset : 0); break;
	}
}

typedef struct {
  float  sow;                   /* s texture ordinate (s over w) */
  float  tow;                   /* t texture ordinate (t over w) */  
  float  oow;                   /* 1/w (used mipmapping - really 0xfff/w) */
}  GrTmuVertex;

typedef struct {
  float x, y;         /* X and Y in screen space */
  float ooz;          /* 65535/Z (used for Z-buffering) */
  float oow;          /* 1/W (used for W-buffering, texturing) */
  float r, g, b, a;   /* R, G, B, A [0..255.0] */
  float z;            /* Z is ignored */
  GrTmuVertex  tmuvtx[GLIDE_NUM_TMU];
} GrVertex;

FX_ENTRY void FX_CALL 
grDrawVertexArray(FxU32 mode, FxU32 Count, void *pointers)
{
	DWORD i;
	switch (mode) {
	case GR_POLYGON:
	case GR_TRIANGLES:
		switch (g_gr_state.origin_location) {
		case GR_ORIGIN_UPPER_LEFT:
			{
				LPD3DTLVERTEX ptlv = tlv;
				GrVertex* pvlist = (GrVertex*)pointers;
				for (i = 0; i < Count; i++) {
					grsubSetTLVertexFromGrVertex(ptlv++, pvlist++);
				}
				break;
			}
		case GR_ORIGIN_LOWER_LEFT:
			{
				LPD3DTLVERTEX ptlv = &tlv[Count];
				GrVertex* pvlist = (GrVertex*)pointers;
				for (i = 0; i < Count; i++) {
					grsubSetTLVertexFromGrVertex(--ptlv, pvlist++);
				}
				break;
			}
		}

		// Draw
		grsubPreRender();
		g_gr_pd3ddevice->DrawPrimitive(GLID3D_POLYGON_D3DPT, D3DFVF_TLVERTEX, tlv, Count, D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);
		grsubPostRender();
		DEBUG(g_gr_nTrianglePerFrame += Count / 3);
		break;
	case GR_TRIANGLE_STRIP:
		switch (g_gr_state.origin_location) {
		case GR_ORIGIN_UPPER_LEFT:
			{
				LPD3DTLVERTEX ptlv = tlv;
				GrVertex* pvlist = (GrVertex*)pointers;
				for (i = 0; i < Count; i++) {
					grsubSetTLVertexFromGrVertex(ptlv++, pvlist++);
				}
				break;
			}
		case GR_ORIGIN_LOWER_LEFT:
			{
				LPD3DTLVERTEX ptlv = &tlv[Count];
				GrVertex* pvlist = (GrVertex*)pointers;
				for (i = 0; i < Count; i++) {
					grsubSetTLVertexFromGrVertex(--ptlv, pvlist++);
				}
				break;
			}
			break;
		}

		// Draw
		grsubPreRender();
		g_gr_pd3ddevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, D3DFVF_TLVERTEX, tlv, Count, D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);
		grsubPostRender();
		DEBUG(g_gr_nTrianglePerFrame += Count / 3);
		break;
	case GR_TRIANGLE_FAN:
		switch (g_gr_state.origin_location) {
		case GR_ORIGIN_UPPER_LEFT:
			{
				LPD3DTLVERTEX ptlv = tlv;
				GrVertex* pvlist = (GrVertex*)pointers;
				for (i = 0; i < Count; i++) {
					grsubSetTLVertexFromGrVertex(ptlv++, pvlist++);
				}
				break;
			}
		case GR_ORIGIN_LOWER_LEFT:
			{
				LPD3DTLVERTEX ptlv = &tlv[Count];
				GrVertex* pvlist = (GrVertex*)pointers;
				for (i = 0; i < Count; i++) {
					grsubSetTLVertexFromGrVertex(--ptlv, pvlist++);
				}
				break;
			}
			break;
		}

		// Draw
		grsubPreRender();
		g_gr_pd3ddevice->DrawPrimitive(D3DPT_TRIANGLEFAN, D3DFVF_TLVERTEX, tlv, Count, D3DDP_DONOTUPDATEEXTENTS | GLID3D_D3DDP_DONOTCLIP);
		grsubPostRender();
		DEBUG(g_gr_nTrianglePerFrame += Count / 3);
		break;
	case GR_TRIANGLE_STRIP_CONTINUE:
	case GR_TRIANGLE_FAN_CONTINUE:
		_exit(0);
		break;
	}
}

FX_ENTRY void FX_CALL 
grDrawVertexArrayContiguous(FxU32 mode, FxU32 Count, void *pointers, FxU32 stride)
{
}

/////////////////////////////////////////////////////////////////////////////
